@@ -1,10 +1,5 @@
 ; extends
 
-(expression_statement
- ((string) @_var @variable)
- (#match? @_var "^[\"']{3}[%]{2}.*[%]{2}[\"']{3}$")
-)
-
 ; it can be # %% [markdown] or # %% [md]
 ((
   (comment) @_mdcomment
