@@ -1,4 +1,5 @@
 #!/use/bin/env python3
+# ruff: noqa: T201
 from __future__ import annotations
 
 import argparse
@@ -12,6 +13,9 @@ import sys
 import tempfile
 import time
 import traceback
+from collections.abc import Sequence
+from datetime import datetime, timezone
+from os import PathLike
 from pathlib import Path
 from urllib.parse import urlparse
 
@@ -22,8 +26,12 @@ import psutil
 import verboselogs
 from git.exc import InvalidGitRepositoryError, NoSuchPathError
 from persistqueue.exceptions import Empty
+from pynvim import Nvim
 from selenium import webdriver
 from selenium.common.exceptions import WebDriverException
+from selenium.webdriver.firefox.options import Options
+from selenium.webdriver.firefox.service import Service
+from selenium.webdriver.remote.webdriver import WebDriver
 from selenium.webdriver.support import expected_conditions as EC
 from selenium.webdriver.support.wait import WebDriverWait
 
@@ -37,14 +45,16 @@ from ..pynvim_helpers import attach_and_init
 
 logger = verboselogs.VerboseLogger(__name__)
 
-SOURCE_DIR = Path(os.path.dirname(os.path.abspath(__file__)))
-
 
 def webdriver_firefox(
-    profiles_ini_path="~/.mozilla/firefox/profiles.ini", profile_name=None
+    profiles_ini_path: str | PathLike | None = "~/.mozilla/firefox/profiles.ini",
+    profile_name: str | None = None,
 ):
     """
+    Get a Firefox webdriver with a specific profile.
+
     profiles.ini path is used to remember the last session (password, etc.)
+
     Args:
         profiles_ini_path: Path to profiles.ini
         profile_name: Profile name in profiles.ini. If None, use the default profile.
@@ -89,11 +99,14 @@ def webdriver_firefox(
 
     logger.info(f"Using firefox profile: {profile_path}")
 
-    profile = webdriver.FirefoxProfile(profile_path)
-    profile.set_preference("browser.link.open_newwindow", 3)
-    profile.set_preference("browser.link.open_newwindow.restriction", 0)
+    options = Options()
+    options.profile = webdriver.FirefoxProfile(profile_path)
+    options.set_preference("browser.link.open_newwindow", 3)
+    options.set_preference("browser.link.open_newwindow.restriction", 0)
     # profile.setAlwaysLoadNoFocusLib(True);
-    return webdriver.Firefox(profile, service_log_path=os.path.devnull)
+
+    service = Service(log_path=os.path.devnull)
+    return webdriver.Firefox(options=options, service=service)
 
 
 # def webdriver_safari():
@@ -117,7 +130,7 @@ def get_parser():
     )
     parser.add_argument(
         "--notebook_URL",
-        default="localhost:8888",
+        default="localhost:8888/nbclassic",
         help="Jupyter Notebook URL to open from the Selenium browser",
     )
     parser.add_argument(
@@ -166,7 +179,8 @@ def get_parser():
         nargs="+",
         default=["jupyter"],
         help="Command to start Jupyter Notebook (but without notebook).\n"
-        "To use conda env, use `--jupyter_command conda run ' --no-capture-output' ' -n' base jupyter`. Notice the space before the dash.\n"
+        "To use conda env, use `--jupyter_command conda run ' --no-capture-output' ' -n' base jupyter`. "
+        "Notice the space before the dash.\n"
         "It is used only when the --notebook_URL is localhost, and is not running.",
     )
     parser.add_argument(
@@ -174,7 +188,7 @@ def get_parser():
         type=str,
         help="When jupyter notebook has started using --jupyter_command, the root dir will be this.\n"
         "If None, open at a git dir of nvim's buffer path and still navigate to the buffer dir.\n"
-        "(e.g. localhost:8888/tree/path/to/buffer)",
+        "(e.g. localhost:8888/nbclassic/tree/path/to/buffer)",
     )
     parser.add_argument(
         "--no_auto_close_tab",
@@ -208,7 +222,8 @@ def start_if_running_else_clear(args, q: persistqueue.UniqueQ):
     else:
         if args.attach_only:
             logger.error(
-                "Jupynium is not running. Remove --attach_only option to start a new instance."
+                "Jupynium is not running. "
+                "Remove --attach_only option to start a new instance."
             )
             return 1
 
@@ -225,6 +240,7 @@ def start_if_running_else_clear(args, q: persistqueue.UniqueQ):
 def number_of_windows_be_list(num_windows: list[int]):
     """
     An expectation for the number of windows to be one of the listed values.
+
     Slightly modified from EC.number_of_windows_to_be(num_windows).
     """
 
@@ -235,10 +251,10 @@ def number_of_windows_be_list(num_windows: list[int]):
 
 
 def attach_new_neovim(
-    driver,
-    new_args,
+    driver: WebDriver,
+    new_args: argparse.Namespace,
     nvims: dict[str, NvimInfo],
-    URL_to_home_windows: dict[str, str],
+    url_to_home_windows: dict[str, str],
 ):
     logger.info(f"New nvim wants to attach: {new_args}")
     if new_args.nvim_listen_addr in nvims:
@@ -246,8 +262,8 @@ def attach_new_neovim(
     else:
         try:
             nvim = attach_and_init(new_args.nvim_listen_addr)
-            if new_args.notebook_URL in URL_to_home_windows.keys():
-                home_window = URL_to_home_windows[new_args.notebook_URL]
+            if new_args.notebook_URL in url_to_home_windows:
+                home_window = url_to_home_windows[new_args.notebook_URL]
             else:
                 prev_num_windows = len(driver.window_handles)
                 driver.switch_to.new_window("tab")
@@ -259,7 +275,7 @@ def attach_new_neovim(
                 sele.wait_until_loaded(driver)
 
                 home_window = driver.current_window_handle
-                URL_to_home_windows[new_args.notebook_URL] = home_window
+                url_to_home_windows[new_args.notebook_URL] = home_window
 
             nvim_info = NvimInfo(
                 nvim, home_window, auto_close_tab=not new_args.no_auto_close_tab
@@ -284,15 +300,17 @@ def generate_notebook_token():
     return secrets.token_urlsafe(16)
 
 
-def exception_no_notebook(notebook_URL, nvim):
+def exception_no_notebook(notebook_url: str, nvim: Nvim | None):
     logger.exception(
-        f"Exception occurred. Are you sure you're running Jupyter Notebook at {notebook_URL}? Use --jupyter_command to specify the command to start Jupyter Notebook."
+        "Exception occurred. "
+        f"Are you sure you're running Jupyter Notebook at {notebook_url}? "
+        "Use --jupyter_command to specify the command to start Jupyter Notebook."
     )
     if nvim is not None:
         nvim.lua.Jupynium_notify.error(
             [
                 "Can't connect to Jupyter Notebook.",
-                f"Are you sure you're running Jupyter Notebook at {notebook_URL}?",
+                f"Are you sure you're running Jupyter Notebook at {notebook_url}?",
                 "Use jupyter_command to specify the command to start Jupyter Notebook.",
             ],
         )
@@ -312,17 +330,20 @@ def kill_child_processes(parent_pid, sig=signal.SIGTERM):
     psutil.wait_procs(children, timeout=3)
 
 
-def kill_notebook_proc(notebook_proc):
+def kill_notebook_proc(notebook_proc: subprocess.Popen | None):
     """
     Kill the notebook process.
-    Used if we opened a Jupyter Notebook server using the --jupyter_command and when no server is running.
+
+    Used if we opened a Jupyter Notebook server using the --jupyter_command
+    and when no server is running.
     """
     if notebook_proc is not None:
         if os.name == "nt":
             # Windows
             os.kill(notebook_proc.pid, signal.CTRL_C_EVENT)
         else:
-            # Need to kill children if the notebook server is started using `conda run` like
+            # Need to kill children if the notebook server is started using
+            # `conda run` like
             # conda run --no-capture-output -n base jupyter notebook
             kill_child_processes(notebook_proc.pid, signal.SIGKILL)
 
@@ -336,8 +357,22 @@ def kill_notebook_proc(notebook_proc):
 
 
 def fallback_open_notebook_server(
-    notebook_port, jupyter_command, notebook_dir, nvim, driver
+    notebook_port: int,
+    notebook_url_path: str,
+    jupyter_command: Sequence[str],
+    notebook_dir: str | PathLike | None,
+    nvim: Nvim | None,
+    driver: WebDriver,
 ):
+    """
+    After firefox failing to try to connect to Notebook, open the Notebook server and try again.
+
+    Args:
+        notebook_url_path: e.g. "/nbclassic"
+
+    Returns:
+        notebook_proc: subprocess.Popen object
+    """
     # Fallback: if the URL is localhost and if selenium can't connect,
     # open the Jupyter Notebook server and even start syncing.
     rel_dir = ""
@@ -347,7 +382,7 @@ def fallback_open_notebook_server(
         if nvim is not None:
             # Root dir of the notebook is either the buffer's dir or the git dir.
             buffer_path = str(nvim.eval("expand('%:p')"))
-            buffer_dir = os.path.dirname(buffer_path)
+            buffer_dir = Path(buffer_path).parent
             try:
                 repo = git.Repo(buffer_dir, search_parent_directories=True)
                 notebook_dir = repo.working_tree_dir
@@ -363,56 +398,90 @@ def fallback_open_notebook_server(
         "--port",
         str(notebook_port),
         "--no-browser",
-        f"--NotebookApp.token",
+        "--NotebookApp.token",
         notebook_token,
+        "--NotebookApp.show_banner=False",
     ]
 
     if notebook_dir is not None:
         # notebook_args += [f"--ServerApp.root_dir={root_dir}"]
         notebook_args += ["--NotebookApp.notebook_dir", notebook_dir]
 
+    notebook_proc = None
     try:
         # strip commands because we need to escape args with dashes.
         # e.g. --jupyter_command conda run ' --no-capture-output' ' -n' env_name jupyter
 
         jupyter_command = [command.strip() for command in jupyter_command]
-        jupyter_command[0] = os.path.expanduser(jupyter_command[0])
+        jupyter_command[0] = str(Path(jupyter_command[0]).expanduser())
 
         jupyter_stdout = tempfile.NamedTemporaryFile()
         logger.info(f"Writing Jupyter Notebook server log to: {jupyter_stdout.name}")
         notebook_proc = subprocess.Popen(
-            jupyter_command + notebook_args, stdout=jupyter_stdout
+            jupyter_command + notebook_args,
+            stdout=jupyter_stdout,
+            stderr=subprocess.STDOUT,
         )
     except FileNotFoundError:
         # Command doesn't exist
-        exception_no_notebook(f"localhost:{notebook_port}", nvim)
+        exception_no_notebook(f"localhost:{notebook_port}{notebook_url_path}", nvim)
+
+    assert notebook_proc is not None
 
     time.sleep(1)
     for _ in range(20):
         try:
             driver.get(
-                f"localhost:{notebook_port}/tree/{rel_dir}?token={notebook_token}"
+                f"localhost:{notebook_port}{notebook_url_path}/tree/{rel_dir}?token={notebook_token}"
             )
             break
         except WebDriverException:
             poll = notebook_proc.poll()
             if poll is not None:
                 # Process finished
-                exception_no_notebook(f"localhost:{notebook_port}", nvim)
+                exception_no_notebook(
+                    f"localhost:{notebook_port}{notebook_url_path}", nvim
+                )
 
         time.sleep(0.3)
     else:
-        # Process still running but timeout for connecting to notebook. Maybe wrong command?
+        # Process still running but timeout for connecting to notebook.
+        # Maybe wrong command?
         kill_notebook_proc(notebook_proc)
-        exception_no_notebook(f"localhost:{notebook_port}", nvim)
+        exception_no_notebook(f"localhost:{notebook_port}{notebook_url_path}", nvim)
     return notebook_proc
 
 
-# flake8: noqa: C901
 def main():
-    coloredlogs.install(
-        fmt="%(name)s: %(lineno)4d - %(levelname)s - %(message)s", level=logging.INFO
+    # Initialise with NOTSET level and null device, and add stream handler separately.
+    # This way, the root logging level is NOTSET (log all),
+    # and we can customise each handler's behaviour.
+    # If we set the level during the initialisation, it will affect to ALL streams,
+    # so the file stream cannot be more verbose (lower level) than the console stream.
+    coloredlogs.install(fmt="", level=logging.NOTSET, stream=open(os.devnull, "w"))  # noqa: SIM115
+
+    console_handler = logging.StreamHandler()
+    console_handler.setLevel(logging.INFO)
+    console_format = coloredlogs.ColoredFormatter(
+        "%(name)s: %(lineno)4d - %(levelname)s - %(message)s"
     )
+    console_handler.setFormatter(console_format)
+
+    tmp_log_dir = Path(tempfile.gettempdir()) / "jupynium" / "logs"
+    tmp_log_dir.mkdir(parents=True, exist_ok=True)
+    log_path = tmp_log_dir / f"{datetime.now(tz=timezone.utc):%Y-%m-%d_%H-%M-%S}.log"
+
+    f_handler = logging.FileHandler(log_path)
+    f_handler.setLevel(logging.INFO)
+    f_format = logging.Formatter(
+        "%(asctime)s - %(name)s: %(lineno)4d - %(levelname)s - %(message)s"
+    )
+    f_handler.setFormatter(f_format)
+
+    # Add handlers to the logger
+    root_logger = logging.getLogger()
+    root_logger.addHandler(console_handler)
+    root_logger.addHandler(f_handler)
 
     parser = get_parser()
     args = parser.parse_args()
@@ -453,17 +522,22 @@ def main():
             args.firefox_profiles_ini_path, args.firefox_profile_name
         ) as driver:
             # Initial number of windows when launching browser
-            ini_num_windows = len(driver.window_handles)
+            init_num_windows = len(driver.window_handles)
             try:
                 driver.get(args.notebook_URL)
             except WebDriverException:
-                notebook_URL = args.notebook_URL
+                notebook_url = args.notebook_URL
                 if "://" not in args.notebook_URL:
-                    notebook_URL = "http://" + notebook_URL
-                url = urlparse(notebook_URL)
+                    notebook_url = "http://" + notebook_url
+                url = urlparse(notebook_url)
                 if url.port is not None and url.hostname in ["localhost", "127.0.0.1"]:
                     notebook_proc = fallback_open_notebook_server(
-                        url.port, args.jupyter_command, args.notebook_dir, nvim, driver
+                        url.port,
+                        url.path,
+                        args.jupyter_command,
+                        args.notebook_dir,
+                        nvim,
+                        driver,
                     )
 
                 else:
@@ -473,16 +547,18 @@ def main():
             # Wait for the notebook to load
             driver_wait = WebDriverWait(driver, 10)
             # Acceptable number of windows is either:
-            # - Initial number of windows, for regular case where jupynium handles initally focused tab
-            # - Initial number of windows + 1, if an extension automatically opens a new tab
+            # - Initial number of windows, for regular case where jupynium handles
+            # initally focused tab
+            # - Initial number of windows + 1, if an extension automatically opens
+            # a new tab
             # Ref: https://github.com/kiyoon/jupynium.nvim/issues/59
-            accept_num_windows = [ini_num_windows, ini_num_windows + 1]
+            accept_num_windows = [init_num_windows, init_num_windows + 1]
             driver_wait.until(number_of_windows_be_list(accept_num_windows))
             sele.wait_until_loaded(driver)
 
             home_window = driver.current_window_handle
 
-            URL_to_home_windows = {args.notebook_URL: home_window}
+            url_to_home_windows = {args.notebook_URL: home_window}
             if args.nvim_listen_addr is not None and nvim is not None:
                 nvims = {
                     args.nvim_listen_addr: NvimInfo(
@@ -491,7 +567,9 @@ def main():
                 }
             else:
                 logger.info(
-                    "No nvim attached. Waiting for nvim to attach. Run jupynium --nvim_listen_addr /tmp/example (use `:echo v:servername` of nvim)"
+                    "No nvim attached. Waiting for nvim to attach. "
+                    "Run jupynium --nvim_listen_addr /tmp/example "
+                    "(use `:echo v:servername` of nvim)"
                 )
 
             while not sele.is_browser_disconnected(driver):
@@ -505,7 +583,8 @@ def main():
                             del_list.append((nvim_listen_addr, None))
                         except Exception:
                             logger.exception(
-                                "Uncaught exception occurred while processing events. Detaching nvim."
+                                "Uncaught exception occurred while processing events. "
+                                "Detaching nvim."
                             )
                             del_list.append((nvim_listen_addr, None))
 
@@ -523,11 +602,11 @@ def main():
 
                     # Check if a new newvim instance wants to attach to this server.
                     try:
-                        new_args = q.get(block=False)
+                        new_args: argparse.Namespace = q.get(block=False)  # type: ignore
                     except Empty:
                         pass
                     else:
-                        attach_new_neovim(driver, new_args, nvims, URL_to_home_windows)
+                        attach_new_neovim(driver, new_args, nvims, url_to_home_windows)
 
                     time.sleep(args.sleep_time_idle)
                 except WebDriverException:
