@@ -17,6 +17,7 @@ if Jupynium_syncing_bufs ~= nil then
 end
 
 Jupynium_syncing_bufs = {} -- key = bufnr, value = 1
+Jupynium_bufs_attached = {} -- key = bufnr, value = 1
 
 function Jupynium_reset_channel()
   vim.g.jupynium_channel_id = -1
@@ -75,7 +76,12 @@ function Jupynium_rpcnotify(event, buf, ensure_syncing, ...)
   rpc(vim.rpcnotify, event, buf, ...)
 end
 
--- block until jupynium responds to the message
+---block until jupynium responds to the message
+---@param event string
+---@param buf integer?
+---@param ensure_syncing boolean
+---@param ... any
+---@return any
 function Jupynium_rpcrequest(event, buf, ensure_syncing, ...)
   if ensure_syncing then
     if Jupynium_syncing_bufs[buf] == nil then
@@ -92,7 +98,7 @@ end
 --- API: Execute javascript in the browser. It will switch to the correct tab before executing.
 ---@param bufnr integer | nil If given, before executing the code it will switch to the tab of this buffer. Requires syncing in advance.
 ---@param code string Javascript code
----@return boolean, object: Success, response
+---@return boolean, any?: Success, response
 function Jupynium_execute_javascript(bufnr, code)
   local ensure_syncing = true
   if bufnr == nil then
@@ -182,7 +188,7 @@ end
 ---Start synchronising the buffer with the ipynb file
 ---@param bufnr integer buffer number
 ---@param ipynb_filename string name of the ipynb file
----@param ask boolean whether to ask for confirmation
+---@param ask boolean? whether to ask for confirmation
 function Jupynium_start_sync(bufnr, ipynb_filename, ask)
   if bufnr == nil or bufnr == 0 then
     bufnr = vim.api.nvim_get_current_buf()
@@ -202,7 +208,7 @@ function Jupynium_start_sync(bufnr, ipynb_filename, ask)
   local content = vim.api.nvim_buf_get_lines(bufnr, 0, -1, false)
 
   -- Used for choosing the correct kernel
-  local buf_filetype = vim.api.nvim_buf_get_option(bufnr, "filetype")
+  local buf_filetype = vim.bo[bufnr].filetype
   local conda_or_venv_path = vim.env.CONDA_PREFIX or vim.env.VIRTUAL_ENV
 
   local response =
@@ -294,16 +300,25 @@ function Jupynium_start_sync(bufnr, ipynb_filename, ask)
     group = augroup,
   })
 
+  if Jupynium_bufs_attached[bufnr] ~= nil then
+    -- on_lines event handler already attached for this buffer
+    return
+  end
+
   vim.api.nvim_buf_attach(bufnr, false, {
     on_lines = function(_, _, _, start_row, old_end_row, new_end_row, _)
       if Jupynium_syncing_bufs[bufnr] == nil then
-        return
+        -- Detach on_lines event handler
+        Jupynium_bufs_attached[bufnr] = nil
+        return true
       end
 
       local lines = vim.api.nvim_buf_get_lines(bufnr, start_row, new_end_row, false)
       Jupynium_rpcnotify("on_lines", bufnr, true, lines, start_row, old_end_row, new_end_row)
     end,
   })
+
+  Jupynium_bufs_attached[bufnr] = 1
 end
 
 function Jupynium_stop_sync(bufnr)
@@ -394,8 +409,8 @@ function Jupynium_download_ipynb_cmd(args)
 end
 
 function Jupynium_auto_download_ipynb_toggle()
-  vim.g.jupynium_autodownload_ipynb = 1 - vim.g.jupynium_auto_download_ipynb
-  Jupynium_notify.info { "Auto download ipynb is now ", vim.g.jupynium_auto_download_ipynb == 1 and "on" or "off" }
+  vim.g.jupynium_auto_download_ipynb = not vim.g.jupynium_auto_download_ipynb
+  Jupynium_notify.info { "Auto download ipynb is now ", vim.g.jupynium_auto_download_ipynb and "on" or "off" }
 end
 
 function Jupynium_scroll_up(bufnr)
@@ -425,8 +440,8 @@ function Jupynium_scroll_down(bufnr)
 end
 
 function Jupynium_autoscroll_toggle()
-  vim.g.jupynium_autoscroll = 1 - vim.g.jupynium_autoscroll
-  Jupynium_notify.info { "Autoscroll is now ", vim.g.jupynium_autoscroll == 1 and "on" or "off" }
+  vim.g.jupynium_autoscroll_enable = not vim.g.jupynium_autoscroll_enable
+  Jupynium_notify.info { "Autoscroll is now ", vim.g.jupynium_autoscroll and "on" or "off" }
 end
 
 function Jupynium_clear_selected_cells_outputs(bufnr)
@@ -477,11 +492,6 @@ function Jupynium_kernel_change(bufnr, kernel_name)
   Jupynium_rpcnotify("kernel_change", bufnr, true, kernel_name)
 end
 
-function Jupynium_restart_kernel(bufnr)
-  Jupynium_notify.warn { [[Sorry! Command name changed.]], [[Please use :JupyniumKernelRestart]] }
-  return Jupynium_kernel_restart(bufnr)
-end
-
 function Jupynium_kernel_restart(bufnr)
   -- note that the kernel name is different from the display name in the kernel list in Jupyter Notebook.
   if bufnr == nil or bufnr == 0 then
@@ -509,11 +519,6 @@ function Jupynium_kernel_interrupt(bufnr)
   Jupynium_rpcnotify("kernel_interrupt", bufnr, true)
 end
 
-function Jupynium_select_kernel(bufnr)
-  Jupynium_notify.warn { [[Sorry! Command name changed.]], [[Please use :JupyniumKernelSelect]] }
-  return Jupynium_kernel_select(bufnr)
-end
-
 function Jupynium_kernel_select(bufnr)
   -- note that the kernel name is different from the display name in the kernel list in Jupyter Notebook.
   if bufnr == nil or bufnr == 0 then
@@ -572,7 +577,7 @@ end
 ---@param bufnr integer
 ---@param code_line string
 ---@param col integer 0-indexed
----@return table | nil
+---@return table?
 function Jupynium_kernel_inspect(bufnr, code_line, col)
   if bufnr == nil or bufnr == 0 then
     bufnr = vim.api.nvim_get_current_buf()
@@ -611,29 +616,53 @@ function Jupynium_kernel_hover(bufnr)
   elseif inspect.found == false then
     out = "No information from kernel"
   elseif inspect.found == true then
-    -- Strip ANSI Escape code: https://stackoverflow.com/a/55324681
-    -- The above regexes do the following:
-    -- 1. \x1b is the escape character
-    -- 2. %[%d+; is the ANSI escape code for a digit color
-    -- and so on
-    out = inspect.data["text/plain"]
-      :gsub("\x1b%[%d+;%d+;%d+;%d+;%d+m", "")
-      :gsub("\x1b%[%d+;%d+;%d+;%d+m", "")
-      :gsub("\x1b%[%d+;%d+;%d+m", "")
-      :gsub("\x1b%[%d+;%d+m", "")
-      :gsub("\x1b%[%d+m", "")
-    -- The following regex convert ansi code for tab
-    -- out = out:gsub("\x1b%[H", "\t")
+    local sections = vim.split(inspect.data["text/plain"], "\x1b%[0;31m")
+    for _, section in ipairs(sections) do
+      section = section
+        -- Strip ANSI Escape code: https://stackoverflow.com/a/55324681
+        -- \x1b is the escape character
+        -- %[%d+; is the ANSI escape code for a digit color
+        :gsub(
+          "\x1b%[%d+;%d+;%d+;%d+;%d+m",
+          ""
+        )
+        :gsub("\x1b%[%d+;%d+;%d+;%d+m", "")
+        :gsub("\x1b%[%d+;%d+;%d+m", "")
+        :gsub("\x1b%[%d+;%d+m", "")
+        :gsub("\x1b%[%d+m", "")
+        :gsub("\x1b%[H", "\t")
+        -- Groups: name, 0 or more new line, content till end
+        -- TODO: Fix for non-python kernel
+        :gsub(
+          "^(Call signature):(%s*)(.-)\n$",
+          "```python\n%3 # %1\n```"
+        )
+        :gsub("^(Init signature):(%s*)(.-)\n$", "```python\n%3 # %1\n```")
+        :gsub("^(Signature):(%s*)(.-)\n$", "```python\n%3 # %1\n```")
+        :gsub("^(String form):(%s*)(.-)\n$", "```python\n%3 # %1\n```")
+        :gsub("^(Docstring):(%s*)(.-)$", "\n---\n```rst\n%3\n```")
+        :gsub("^(Class docstring):(%s*)(.-)$", "\n---\n```rst\n%3\n```")
+        :gsub("^(File):(%s*)(.-)\n$", "*%1*: `%3`\n")
+        :gsub("^(Type):(%s*)(.-)\n$", "*%1*: %3\n")
+        :gsub("^(Length):(%s*)(.-)\n$", "*%1*: %3\n")
+        :gsub("^(Subclasses):(%s*)(.-)\n$", "*%1*: %3\n")
+      if section:match "%S" ~= nil and section:match "%S" ~= "" then
+        -- Only add non-empty section
+        out = out .. section
+      end
+    end
   end
 
-  local lines = {}
-  for line in vim.gsplit(out, "\n") do
-    table.insert(lines, line)
+  local markdown_lines = vim.lsp.util.convert_input_to_markdown_lines(out)
+  markdown_lines = vim.lsp.util.trim_empty_lines(markdown_lines)
+
+  local opts = { max_width = 84 }
+  local ok, options = pcall(require, "jupynium.options")
+  if ok then
+    opts = vim.tbl_extend("force", opts, options.opts.kernel_hover.floating_win_opts)
   end
 
-  vim.lsp.util.open_floating_preview(lines, "markdown", {
-    max_width = 84,
-  })
+  vim.lsp.util.open_floating_preview(markdown_lines, "markdown", opts)
 end
 
 local function get_memory_addr(t)
@@ -645,7 +674,7 @@ end
 ---@param code_line string
 ---@param col integer 0-indexed
 ---@param callback function nvim-cmp complete callback.
----@return table | nil
+---@return table?
 function Jupynium_kernel_complete_async(bufnr, code_line, col, callback)
   if bufnr == nil or bufnr == 0 then
     bufnr = vim.api.nvim_get_current_buf()
@@ -668,3 +697,42 @@ function Jupynium_kernel_complete_async(bufnr, code_line, col, callback)
 
   Jupynium_rpcnotify("kernel_complete_async", bufnr, true, code_line, col, callback_id)
 end
+
+function Jupynium_get_kernel_connect_shcmd(bufnr, hostname)
+  if bufnr == nil or bufnr == 0 then
+    bufnr = vim.api.nvim_get_current_buf()
+  end
+
+  local kernel_id = nil
+  if Jupynium_syncing_bufs[bufnr] ~= nil then
+    kernel_id = Jupynium_rpcrequest("kernel_connect_info", bufnr, true)
+  end
+  if kernel_id == nil then
+    kernel_id = ""
+  end
+  local jupyter_command = "jupyter"
+  local ok, options = pcall(require, "jupynium.options")
+  if ok then
+    if type(options.opts.jupyter_command) == "string" then
+      jupyter_command = options.opts.jupyter_command
+    elseif type(options.opts.jupyter_command) == "table" then
+      jupyter_command = table.concat(options.opts.jupyter_command, " ")
+    else
+      Jupynium_notify.error { "Invalid jupyter_command type." }
+    end
+  end
+  if hostname ~= "" then
+    jupyter_command = "ssh " .. hostname .. " -t " .. jupyter_command
+  end
+  Jupynium_notify.info { "Connecting to kernel " .. kernel_id }
+  local cmd = jupyter_command .. " console --existing " .. kernel_id
+  return cmd
+end
+
+function Jupynium_kernel_connect_cmd(args)
+  local hostname = args.args
+  local buf = vim.api.nvim_get_current_buf()
+  local cmd = Jupynium_get_kernel_connect_shcmd(buf, hostname)
+  vim.cmd([[split | terminal ]] .. cmd)
+  vim.cmd [[normal! G]]
+end
